#!/usr/bin/perl -w
#
# perlTiger v1.0
# Last modified May 1, 2020 14:10
# (c) Olaf R.P. Bininda-Emonds
#
# Purpose: Implements the TIGER method of Cummings and McInerney (2011) for
#		   determining the relative rates of evolution for each position in a
#		   DNA alignment
#
# Input: An aligned DNA data set in any of fasta, nexus, or phylip formats
#
# Output:
#	1) A tab-delimited text file listing the position, its TIGER rate, the
#      number of informative positions, and optionally the bin number of its
#      rate 
#	2) A tab-delimited text file listing bin information as the bin number, its
#      lower rate boundary, its uper rate boundary, the number of positions in
#      the bin, and the positions (optional)
#
# Usage: perlTiger.pl -d <filename> [-b <number>] [-e <symbol>] [-g]
#        [-i <f|n|p>] [-s] [-v] [-h]
#	options: -b <number> = number of rate bins as a natural number (default =
#                          0; off)
#            -d <filename> = file containing raw DNA sequence information
#			 -e <symbol> = symbols to be excluded from rate calculations
#                          (default includes N, ?, *, and X; flag should be
#                          invoked multiple times for additional symbols)
#			 -g = include flanking gaps in calculations
#            -i <f|n|p> = format of sequence file (fasta (f), nexus (n), or
#					      phylip (p); default = autodetect
#	         -s = enable strict interpretation of IUPAC symbols (default =
#                 resolved)
#	         -v = verbose output
#            -h = print this message and quit
#
# Citation: Cummins, C.A. and J.O. McInerney, J.O. 2011. A method for inferring
# (TIGER)   the rate of evolution of homologous characters that can potentially
#           improve phylogenetic inference, resolve deep divergence and correct
#           systematic biases. Systematic Biology 60(6): 833-844.;
# Citation:    Bininda-Emonds, O.R.P. In prep. 18S rRNA variability maps reveal
# (perlTiger)  highly divergent, conserved motifs within Rotifera.

use strict;

# Set default parameters
	# I/O variables
		my $iupacMatch = "resolved";	# Options are strict / resolved
		my $flankingGaps = "X";
		my @ignoreCodes;
			push @ignoreCodes, "N", "?", "*", "X";	# IUPAC codes here too?
			my %ignoreCode;
		my $numBins = 0;

	# Data set variables
		my %tigerRate;
		
		my $inputType = "";	# Options are "fasta", "GenBank", "nexus", "phylip", and "Se-Al"
		my ($readFile, $dataSource);
 		my $seqType = "DNA";
		my $globalGenCode = 0;
			my $globalOverride = 0;
		my $orientation = "asis";

		my (@accNum, %nameLabel, %sequence, %geneticCode, %accPresent);
		my (@charsetList, %charsetStart, %charsetEnd);
		my (%deletedSeq, %finalSeq);
		my $seqCount;
		my $ntax;
		my $nameClean = 1;
		my $skipDuplicates = 1;
		my (%specimenCount, @rootNames);

	# seqRead variables
 		my $seqOrder = "alphabetical";	# Options are "alphabetical" (default), "input", "length", "five", "three", "guideTree" and "custom"
			my ($orderStart, $orderStop);
 			my $aaSort = 0;
 			my $guideTreeFile;
 		my $seqSplit = 0;
 			my $charsetFile;
 			my %partitionPositions;
 		my $subDivide = 1;
 		my $jackknife = 0;
 		my $gapChange = 1;
 		my $ambigChange = 0;
 		my $translateSeqs = 0;
		my $haploTyping = 0;
			my $haploFile;
		my $flankGap = 0;
		my $ryCoding;
		my $noRNA = 1;
		my $consensus = 0;

	# Translation variables and genetic codes
		my %transTable = ('1' => 'standard',
						  '2' => 'vertebrate mitochondrial',
						  '3' => 'yeast mitochondrial',
						  '4' => 'mold, protozoan and colenterate mitochondrial and mycoplasam/spiroplasma',
						  '5' => 'invertebrate mitochondrial',
						  '6' => 'ciliate, dasycladacean and hexamita nuclear',
						  '9' => 'echinoderm mitochondrial',
						  '10' => 'euplotid nuclear',
						  '11' => 'bacterial and plant plastid',
						  '12' => 'alternative yeast nuclear',
						  '13' => 'ascidian mitochondrial',
						  '14' => 'alternative flatworm mitochondrial',
						  '15' => 'Blepharisma nuclear',
						  '16' => 'chlorophycean mitochondrial',
						  '21' => 'trematode mitochondrial',
						  '22' => 'Scenedesmus obliquus mitochondrial',
						  '23' => 'Thraustochytrium mitochondrial');
		my %gb2seal = ('1' => '0',
					   '2' => '1',
					   '3' => '2',
					   '4' => '3',
					   '4' => '4',
					   '5' => '5',
					   '6' => '6',
					   '9' => '7',
					   '10' => '8',
					   '11' => '9',
					   '12' => '10',
					   '13' => '11',
					   '14' => '12',
					   '15' => '13',
					   '16' => '1',
					   '21' => '1',
					   '22' => '1',
					   '23' => '1');
		my %seal2gb = ('0' => '1',
					   '1' => '2',
					   '2' => '3',
					   '3' => '4',
					   '4' => '4',
					   '5' => '5',
					   '6' => '6',
					   '7' => '9',
					   '8' => '10',
					   '9' => '11',
					   '10' => '12',
					   '11' => '13',
					   '12' => '14',
					   '13' => '15');
		my %genCodes;
			$genCodes{$_} = 1 foreach (qw(0 1 2 3 4 5 6 9 10 11 12 13 14 15 16 21 22 23));
		my %DNAtoAA;
		my @ambigList = qw(A C G T M R W S Y K V H D B N);
			my %informBP;
				$informBP{$_} = 1 foreach (@ambigList);
				$informBP{"N"} = $informBP{"?"} = $informBP{"-"} = 0;
		my %ambigCode = ('A' => 'A', 'C' => 'C', 'G' => 'G', 'T' => 'T',
						'AC' => 'M', 'AG' => 'R', 'AT' => 'W', 'CG' => 'S', 'CT' => 'Y', 'GT' => 'K',
						'ACG' => 'V', 'ACT' => 'H', 'AGT' => 'D', 'CGT' => 'B', 'ACGT' => 'N');
		my (%constitNTlist);
			while (my ($nt, $code) = each %ambigCode)	# Where $nt = key and $code = value
				{
				push @{$constitNTlist{$code}}, $_ foreach (split("",$nt));
				}
	
	# binMaker variables
		my $binMinimum = 0;
		my $binMaximum = 1;
		my $binWidth;
		my (%lowerBin, %upperBin);
		my %binCount;
			$binCount{"-1"} = $binCount{"0"} = 0;
		my %binCharacters;

	# Miscellaneous variables
		my @errorList;
		my $quiet = 0;
		my $verbose = 0;
		my $debug = 0;
		my $perlScript = "perlTiger.pl";
			my $callString = join(' ', $perlScript, @ARGV);
		my $version = "1.0";
		my $usage = "Usage: $perlScript -d <filename> [-b <number>] [-e <symbol>] [-g]\n";
			$usage .= "       [-i <f|n|p|s>] [-s] [-v] [-h]";
		my $options .= "Options: -b <number> = number of rate bins as a natural number (default =\n";
			$options .= "                       0; off)\n";
			$options .= "         -d filename = file containing raw DNA sequence information\n";
			$options .= "         -e <symbol> = symbols to be excluded from rate calculations\n";
			$options .= "                       (default includes N, ?, *, and X; flag should be\n";
			$options .= "                       invoked multiple times for additional symbols)\n";
			$options .= "         -g = include flanking gaps in calculations\n";
			$options .= "         -i <f|n|p> = format of sequence file (fasta (f), nexus (n), or\n";
			$options .= "                      phylip (p); default = autodetect\n";
			$options .= "         -s = enable strict interpretation of IUPAC symbols (default =\n";
			$options .= "              resolved\n";
			$options .= "         -h = print this message and quit\n";
			$options .= "         -v = verbose output\n";
			$options .= "\nCitation: Cummins, C.A. and J.O. McInerney, J.O. 2011. A method for inferring\n";
			$options .= "(TIGER)   the rate of evolution of homologous characters that can potentially\n";
			$options .= "          improve phylogenetic inference, resolve deep divergence and correct\n";
			$options .= "          systematic biases. Systematic Biology 60(6): 833-844.\n";
			$options .= "\nCitation:    Bininda-Emonds, O.R.P. In prep. 18S rRNA variability maps reveal\n";
			$options .= "(perlTiger)  highly divergent, conserved motifs within Rotifera.";

# Process flags
	if (join(' ', @ARGV) =~ /\s-h/ or $ARGV[0] =~ /^-h/)	# Print help screen
		{
		print "$usage\n";
		print "Version: $version\n";
		print "$options\n";
		exit(0);
		}
	else	# Process switches
		{
		my ($switch, $arg);
		until (not @ARGV)
			{
			$_ = shift(@ARGV);
				if ($_ =~ /^-/)
					{
					$switch = $_;
					undef $arg;
					}
				else
					{
					$arg = $_;
					}

			if ($switch eq "-b")
				{
				next unless (defined $arg);
				
				if ($arg =~ /\D/ or $arg == 0)
					{
					push @errorList, "Number of rate bins must be a natural number.\n";
					}
				else
					{
					$numBins = $arg;
					&binMaker;
					}
				}		
			elsif ($switch eq "-d")
				{
				next unless (defined $arg);
				
				if (-e $arg)
					{
					$readFile = $arg;
					}
				else
					{
					push @errorList, "ERROR: Data file $arg does not exist.\n";
					}
				}
			elsif ($switch eq "-e")
				{
				next unless (defined $arg);
				
				push @ignoreCodes, $arg;
				}
			elsif ($switch eq "-g")
				{
				$flankingGaps = "";
				}
			elsif ($switch eq "-i")
				{
				next unless (defined $arg);
				if ($arg eq "f")
					{
					$inputType = "fasta";
					}
				elsif ($arg eq "g")
					{
					$inputType = "GenBank";
					}
				elsif ($arg eq "n")
					{
					$inputType = "nexus";
					}
				elsif ($arg eq "p")
					{
					$inputType = "phylip";
					}
				elsif ($arg eq "s")
					{
					$inputType = "Se-Al";
					}
				else
					{
					push @errorList, "Invalid input format: $arg\n";
					}
				}
			elsif ($switch eq "-s")
				{
				$iupacMatch = "strict";
				}
			elsif ($switch eq "-v")
				{
				$verbose = 1;
				}
			elsif ($switch eq "-x")
				{
				$debug = 1;
				$verbose = 1;
				}
			else
				{
				push @errorList, "Unrecognized switch: $switch\n";
				}
			}
		}
		
	# Process entries
		$ignoreCode{$_} = 1 foreach (@ignoreCodes);

	# Check for I/O errors
		push @errorList, "ERROR: Must supply name of file containing sequence data.\n" if (not $readFile);
		
		if (@errorList)
			{
			print "$_" foreach (@errorList);
			print "\n$usage\n";
			print "Version: $version\n";
			exit(1);
			}

		print "Script was called as (" . localtime() . "):\n\t$callString\n";

# Read in and preprocess sequence data
	($dataSource = $readFile) =~ s/\.\w+$//;
		my $ratesFile = $dataSource . "_tigerRates.txt";
			open (RATE, ">$ratesFile") or die "Cannot write TIGER rates to file $ratesFile ($!)\n";

	seqRead($readFile);
	if (not @accNum)
		{
		print "ERROR: Could not read in sequences from file $readFile; skipping to next file\n";
		exit(1);
		}
		
	# Get alignment length and convert flanking gaps if desired
		my $maxLength = 0;
		foreach my $accession (@accNum)
			{
			$maxLength = length($sequence{$accession}) if (length($sequence{$accession}) > $maxLength);
			if (defined $flankingGaps)
				{
				if ($sequence{$accession} =~ /^(\-+)/)
					{
					my $startGap = $1;
						my $startN = $flankingGaps x length($startGap);
					$sequence{$accession} =~s/^$startGap/$startN/;
					}
				if ($sequence{$accession} =~ /(\-+)$/)
					{
					my $endGap = $1;
						my $endN = $flankingGaps x length($endGap);
					$sequence{$accession} =~s/$endGap$/$endN/;
					}
				}
			}

	# Add fitting 3' gaps to make all sequences the same length
		foreach my $accession (@accNum)
			{
			my $missingLength = $maxLength - length($sequence{$accession});
			
			if ($missingLength)
				{
				$sequence{$accession} .= $flankingGaps for(1..$missingLength);
				}
			}

# Calculate set partitions at each position
	print "\nCalculating TIGER rates ...\n";
		my $tigerZero = time;
	
	my (%partitionList, %positionPartitions);
	my %taxonCount;
	my @partitionSymbols = qw(A C G T -);
		push @partitionSymbols, "M", "R", "W", "S", "Y", "K", "V", "H", "D", "B" if ($iupacMatch eq "strict");
#		push @partitionSymbols, "X" if ($debug);
		
	for (my $position = 0; $position < $maxLength; $position++)
		{
		print "Position: $position\n" if ($debug);
		my $taxon = 0;
		$taxonCount{$position} = 0;

		foreach my $accession (@accNum)
			{
			$taxon++;
			my $bp = substr($sequence{$accession}, $position, 1);
			next if (defined $ignoreCode{$bp});
				$taxonCount{$position}++;
				
			if ($iupacMatch eq "strict")
				{
				push @{ $partitionList{$position}{$bp}}, $taxon;
				}
			else
				{
				push @{ $partitionList{$position}{"X"}}, $taxon if ($debug and $bp eq "X");
				foreach my $nt (@{$constitNTlist{$bp}})
					{
					push @{ $partitionList{$position}{$nt}}, $taxon;
					}
				}
			}

		foreach my $symbol (@partitionSymbols)
			{
			print "$symbol\t" if ($debug);
			if (exists $partitionList{$position}{$symbol})
				{
				my $partition = join(',', @{ $partitionList{$position}{$symbol}});
					push @{ $positionPartitions{$position}}, $partition;
				print "$partition\n" if ($debug);
				}
			else
				{
				print "-\n" if ($debug);
				}
			}
		}
	print "\n" if ($debug);

# Calculate TIGER rates
	for (my $i = 0; $i < $maxLength; $i++)
		{
		$tigerRate{$i} = 0;

		my $partitionSet = join('|', @{ $positionPartitions{$i}}) if (exists $positionPartitions{$i});
		print "iPosition: $i\t($partitionSet)\n" if ($debug);

			if (not exists $positionPartitions{$i} or $taxonCount{$i} == 1)	# Position is uninformative
				{
				$tigerRate{$i} = sprintf "%.5f", -1 / 1;
				}
			elsif (scalar(@{ $positionPartitions{$i}}) == 1)	# Position is constant
				{
				$tigerRate{$i} = sprintf "%.5f", 1 / 1;
				}
			elsif (defined $tigerRate{$partitionSet})	# Pattern has been seen before
				{
				$tigerRate{$i} = $tigerRate{$partitionSet};
				}
			else	# Determine partition agreement score
				{
				my $jCount = 0;
				for (my $j = 0; $j < $maxLength; $j++)
					{
					my $pas = 0;
					next if ($i == $j);
					next if (not exists $positionPartitions{$j} or scalar(@{ $positionPartitions{$j}}) == 1);	# Position j is uninformative or constant, respectively
						$jCount++;
			
					printf "\tjPosition: $j (%s)\n", join('|', @{ $positionPartitions{$j}}) if ($debug);
					foreach my $jPartition (@{ $positionPartitions{$j}})
						{
						print "\t\tjPartition: $jPartition\n" if ($debug);
						my @jPartition = split(',', $jPartition);

						foreach my $iPartition (@{ $positionPartitions{$i}})
							{
							print "\t\t\tiPartition: $iPartition\n" if ($debug);
							my @iPartition = split(',', $iPartition);				

							my $subset = 0;
								$subset += subset(\@iPartition, \@jPartition);

								if ($debug)
									{
#									print "\t\t\t\ti: @iPartition\n";
#									print "\t\t\t\tj: @jPartition\n";
#									print "\t\t\t\tintersection: @intersection\n";
									print "\t\t\t\tj is subset of i\n" if ($subset);
									}
							$pas += $subset;
							}
						}
					printf "\n\t\traw PAS{$i},{$j}: $pas (n = %s partitions)\n", scalar(@{ $positionPartitions{$j}}) if ($debug);
					$pas /= scalar(@{ $positionPartitions{$j}});
						print "\t\t\tfinal PAS{$i},{$j}: $pas\n" if ($debug);
					$tigerRate{$i} += $pas;
						if ($debug)
							{
							print "\t\t\tcumulative PAS{$i},{$j}: $tigerRate{$i} (n = $jCount positions)\n" if ($debug);
							printf "\t\t\t\tcurrent rate: %.5f\n\n", $tigerRate{$i} / $jCount;
							}
					}

				print "\traw rate ($i): $tigerRate{$i}\n\n" if ($debug);
				$tigerRate{$i} = sprintf "%.5f", $tigerRate{$i} / $jCount;
					$tigerRate{$partitionSet} = $tigerRate{$i};
				}
		}
		
	for (my $i = 0; $i < $maxLength; $i++)
		{
		printf RATE "%s\t$tigerRate{$i}\t$taxonCount{$i}", $i + 1;
			if ($numBins)
				{
				my $bin = $numBins - int($tigerRate{$i} / $binWidth);
					$bin = -1 if ($tigerRate{$i} == -1);
					$bin = 0 if ($tigerRate{$i} == 1);
				$binCount{$bin}++;
				push @{ $binCharacters{$bin}}, $i+1;

				$bin = "uninf" if ($bin == -1);
				print RATE "\t$bin" if ($numBins);
				}
			print RATE "\n";
		}
	close RATE;

# Print out charsets with bin information as needed
	if ($numBins)
		{
		my $charsetFile = $dataSource . "_tigerBins.txt";
			open (BIN, ">$charsetFile") or die "Cannot write bin information to file $charsetFile ($!)\n";
				print BIN "Number\tLower boundary\tUpper boundary\tNumber of characters\tCharset\n";
			
			for (my $i = -1; $i <= $numBins; $i++)
				{
				if ($i == -1)
					{
					print BIN "uninformative\t\t";
					}
				elsif ($i == 0)
					{
					print BIN "constant\t\t";
					}
				else
					{
					print BIN "$i\t$lowerBin{$i}\t$upperBin{$i}";
					}

				print BIN "\t$binCount{$i}";
				my $list = "";
					$list = listMaker(\@{ $binCharacters{$i}}) if (exists $binCharacters{$i});
					print BIN "\t$list\n";
				}
		close BIN;
		}

printf "\tTime taken: %s seconds ...\n", time - $tigerZero;

exit(0);

### Subroutines used in the program

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks ($!)\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}

sub seqRead
	{
	my $seqFile = shift;
	undef %sequence;

	print "\nReading in sequence data from file $seqFile (type is $inputType) ...\n" if ($inputType and not $quiet);
	setLineBreak($seqFile);
	open (SEQ, "<$seqFile") or die "Cannot open file containing sequences, $seqFile ($!)\n";
		my ($header, $tempAcc, $tempName, $tempSeq);
		my $fastaAcc;
		my (%nexusSpecies, %nexusAcc, $nexusRead, $commentFlag);
		my ($phylipLineCount, $phylipTaxa, $phylipChars, %phylipSeq);
		my $sealCode;
		my ($sealDelFlag, $owner) = (0, 0);
		my ($gbAcc, $gbRead);
		my $macBlock = 0;
		my %accCount;

		while (<SEQ>)
			{
			chomp;
			my $lineRead = $_;
			next unless ($lineRead);
			
			# Autodetect sequence format
				if (not $inputType)
					{
					$inputType = "fasta" if ($lineRead =~ /^>/);
					$inputType = "nexus" if ($lineRead =~ /\#nexus/i);
					$inputType = "phylip" if ($lineRead =~ /^\s*\d+\s+\d+/);
					$inputType = "Se-Al" if ($lineRead =~ /^\s*Database=\{/i);
					$inputType = "GenBank" if ($lineRead =~ /^\s*LOCUS/);
					print "\nReading in sequence data from file $seqFile (type determined to be $inputType) ...\n" if ($inputType);
					}
			
			if ($inputType eq "nexus")
				{
				# Check if charset statement present
					if ($lineRead =~ /charset/i)
						{
						$lineRead =~ s/\s+/ /g;
						$lineRead =~ s/\s\-\s/-/g;
						$lineRead =~ s/;$//;

						my ($charsetName, $charsetBounds) = split('=', $lineRead);
							$charsetName =~ s/charset//i;
								$charsetName =~ s/^\s+//;
								$charsetName =~ s/\s+$//;
								push @charsetList, $charsetName;
								$partitionPositions{$charsetName} = $charsetBounds;
						}
						
				# Otherwise block out MacClade / PAUP blocks
					$macBlock = 1 if ($lineRead =~ /begin macclade;/i or $lineRead =~ /begin paup;/i);
					$macBlock = 0 if ($macBlock and $lineRead =~ /end;/i);
					next if ($macBlock);
						
				# Otherwise only read in data lines or charset statements
					if ($lineRead =~ /^\s*matrix/i)
						{
						$nexusRead = 1;
						next;
						}
					$nexusRead = 0 if ($lineRead =~ /;\s*$/);
					next unless ($nexusRead);
					
					# Remove MacClade sequence lengths
						$lineRead =~ s/\[\d+\]$//;

					$commentFlag = 1 if ($lineRead =~ /\[/);
					if ($lineRead =~ /\]/)
						{
						$commentFlag = 0;
						next;
						}
					next if ($commentFlag);

					next unless ($lineRead =~ /a/i or $lineRead =~ /c/i or $lineRead =~ /g/i or $lineRead =~ /t/i or $lineRead =~ /n/i or $lineRead =~ /\?/ or $lineRead =~ /\-/);

				# Clean up input line
					$lineRead =~ s/^\s+//;
#					$lineRead =~ s/\'//g;

				my @nexusLine = split(/\s+/, $lineRead);
					my $species = shift(@nexusLine);
					my $tempAcc;
						if ($species =~ /^'/)	# Contains non-alphanumerics, possibly including spaces, possibly including accession numbers
							{
							until ($species =~ /'$/)
								{
								$species .= "_" . shift(@nexusLine);
								}
							# Check for accession numbers
								if ($species =~ /(\(?[A-Z]{1,2}_*\d{5,6}\.?\d*\)?)/ or $species =~ /(\(?[A-Z]{2,3}\_\d+[A-Z]\d{2}\)?)$/)
									{
									$tempAcc = $1;
										$tempAcc =~ s/\(//;
										$tempAcc =~ s/\)//;
									$species =~ s/$tempAcc//;
										$species =~ s/\(\)//;
										$species =~ s/^'//;
										$species =~ s/'$//;
									}
							}

						$species =~ s/\s+/_/g;
						$species =~ s/\_+/_/g;
						$species =~ s/\_$//g;
						
						$nexusAcc{$species} = $tempAcc if (defined $tempAcc);

					my $seq = join('', @nexusLine);
						$seq =~ s/\s+//g;
						$seqType = "protein" if ($seq =~ /E/i or $seq =~ /Q/i or $seq =~ /I/i or $seq =~ /L/i or $seq =~ /F/i or $seq =~ /P/i);

				if (not defined $nexusSpecies{$species})
					{
					$nexusSpecies{$species} = 1;
					$seqCount++;
					$nexusAcc{$species} = "tAlign_".$seqCount unless (defined $nexusAcc{$species});
					push @accNum, $nexusAcc{$species};
						$nameLabel{$nexusAcc{$species}} = $species;
						$sequence{$nexusAcc{$species}} = uc($seq);
						$geneticCode{$nexusAcc{$species}} = $globalGenCode;
					}
				else	# Sequences are in interleaved format; append sequence
					{
					$sequence{$nexusAcc{$species}} .= uc($seq);
					}
				}

			if ($inputType eq "fasta")
				{
				if ($lineRead =~/^\s*>/)
					{
					my $species;
					$seqCount++;
					(my $tempSpecies = $lineRead) =~ s/^\s*>//;
					
					# Parse different formats
						if ($tempSpecies =~ /^Mit\.\s+/)	# Entry comes from European RNA project
							{
							$tempSpecies =~ s/^Mit\.\s+//i;	# To fix entries from European RNA project
							my @speciesInfo = split(/\s+/, $tempSpecies);
								$species = join('_', $speciesInfo[0], $speciesInfo[1]);
							if (defined $speciesInfo[2])
								{
								$fastaAcc = $speciesInfo[2];
								$accCount{$fastaAcc}++;
									if ($accCount{$fastaAcc} > 1)
										{
										print "\nWARNING: Accession number $fastaAcc used more than once";
											print "; skipping this and subsequent entries" if ($skipDuplicates);
										print "\n";
										}
								}
							else
								{
								$fastaAcc = "tAlign_".$seqCount;
								$accCount{$fastaAcc}++;
								}
							}
						elsif ($tempSpecies =~ /^(\w+\.\d+\.\d+)/)	# Raw SILVA / ARB RNA project
							{
							$fastaAcc = $1;
							$accCount{$fastaAcc}++;
								if ($accCount{$fastaAcc} > 1)
									{
									print "\nWARNING: Accession number $fastaAcc used more than once";
										print "; skipping this and subsequent entries" if ($skipDuplicates);
									print "\n";
									}
									
							($species = $tempSpecies) =~ s/^\w+\.\d+\.\d+//;
								$species =~ s/^\s+//;
								$species =~ s/\s+/_/g;
							}
						else
							{
							# Parse Seaview specific code
								$tempSpecies =~ s/\s*no comment//;
								if ($tempSpecies =~ /\s+\/transl_table=(\d+)$/)
									{
									$sealCode = $1;
									$tempSpecies =~ s/\s+\/transl_table=\d+$//;
									}
#								if ($tempSpecies =~ /_(\(?[A-Z]{1,2}_*\d{5,6}\.?\d*\)?)$/ or $tempSpecies =~ /_(\(?[A-Z]{2,3}\_\d+[A-Z]\d{2}\)?)$/ or $tempSpecies =~ /_(\(\w+\))$/)
#									{
#									my $tempAcc = $1;
#										$tempSpecies =~ s/$tempAcc//;
#										$tempSpecies =~ s/\(\)//;
#										$tempSpecies =~ s/_$//;
#										$tempSpecies .= "\t$tempAcc";
#									}
									
#							$tempSpecies =~ s/\s+/_/g;
							my @speciesLine = split(/\s+/, $tempSpecies);
							if (($speciesLine[$#speciesLine] =~ /\(?[A-Z]{1,2}_*\d{5,6}\.?\d*\)?/ or $speciesLine[$#speciesLine] =~ /\w+\.\d+\.\d+/ or $speciesLine[$#speciesLine] =~ /\(?[A-Z]{2,3}\-?\d+[A-Z]\d{2}\)?/ or $speciesLine[$#speciesLine] =~ /\(.+\)$/) and scalar(@speciesLine) > 1)	# Check whether last entry is an accession number
								{
								$fastaAcc = pop (@speciesLine);
								$fastaAcc =~ s/^\(//g;
								$fastaAcc =~ s/\)$//g;
								$accCount{$fastaAcc}++;
									if ($accCount{$fastaAcc} > 1)
										{
										print "\nWARNING: Accession number $fastaAcc used more than once";
											if ($skipDuplicates)
												{
												print "; skipping this and subsequent entries\n";
												}
											else
												{
												print "; assigning temporary accession number\n";
												$fastaAcc = "tAlign_" . $seqCount;
												}
										}
								}
							else
								{
								$fastaAcc = "tAlign_".$seqCount;
								$accCount{$fastaAcc}++;
								}
							$species = join('_', @speciesLine);
								$species = "Sequence_".$seqCount if ($species eq "");
							}
					push @accNum, $fastaAcc unless ($accCount{$fastaAcc} > 1 and $skipDuplicates);
						$geneticCode{$fastaAcc} = $globalGenCode;
					($nameLabel{$fastaAcc} = $species) =~ s/\s+/_/;
						$nameLabel{$fastaAcc} =~ s/\_+/_/;
					}
				else
					{
					next if ($accCount{$fastaAcc} > 1 and $skipDuplicates);
					$lineRead =~ s/\s+//g;
						$seqType = "protein" if ($lineRead =~ /E/i or $lineRead =~ /Q/i or $lineRead =~ /I/i or $lineRead =~ /L/i or $lineRead =~ /F/i or $lineRead =~ /P/i);
					$sequence{$fastaAcc} .= uc($lineRead);
					}
				$geneticCode{$fastaAcc} = $sealCode;
				}

			if ($inputType eq "Se-Al")
				{
				my $header;
				$sealDelFlag = 1 if ($lineRead =~ /MCoL/ or $lineRead =~ /^\}\;/);	# Se-Al sometimes places deleted species at end of file; do not read in remainder of file
					next if ($sealDelFlag == 1);
				next unless ($lineRead =~ /NumSites=/ or $lineRead =~ /Owner=/ or $lineRead =~ /Name=/ or $lineRead =~ /Accession=/ or $lineRead =~ /Sequence=/ or $lineRead =~/GeneticCode=/ or $lineRead =~ /Frame=/);
					$lineRead =~ s/^\s+//;

				if ($lineRead =~ /^Owner\s*\=\s*(\d+)/)
					{
					$owner = $1;
					}
				if ($lineRead =~ /^Accession=/ and $owner == 2)
					{
					$seqCount++;
					if ($lineRead =~ /null/ or $lineRead =~ /\"\"/)
						{
						$tempAcc = "tAlign_" . $seqCount;
						$accCount{$tempAcc}++;
						}
					else
						{
						($header, $tempAcc) = split (/=/, $lineRead);
							$tempAcc =~ s/\"//g;
							$tempAcc =~ s/;//g;
							$accCount{$tempAcc}++;
								if ($accCount{$tempAcc} > 1)
									{
									print "\nWARNING: Accession number $tempAcc used more than once";
										if ($skipDuplicates)
											{
											print "; skipping this and subsequent entries\n";
											}
										else
											{
											print "; assigning temporary accession number\n";
											$tempAcc = "tAlign_" . $seqCount;
											}
									print "\n";
									}
						}
					}
				if ($lineRead =~ /^Name=/ and $owner == 2)
					{
					($header, $tempName) = split (/=/, $lineRead);
						$tempName =~ s/\"//g;
						$tempName =~ s/\s*;//g;
						$tempName =~ s/\s+/_/g;
						$tempName =~ s/\_+/_/g;
					}
				if ($lineRead =~ /^GeneticCode=(\d+)/ and $owner == 2)
					{
					$sealCode = $1;
					$geneticCode{$tempAcc} = $seal2gb{$sealCode};
					}
				if ($lineRead =~ /^Sequence=/ and $owner == 2)
					{
					next if ($accCount{$tempAcc} > 1 and $skipDuplicates);
					($header, $tempSeq) = split (/=/, $lineRead);
						$tempSeq =~ s/\"//g;
						$tempSeq =~ s/;//g;
						$tempSeq =~ s/\s+//g;
						next if ($tempSeq eq "null");
						
					push @accNum, $tempAcc unless ($accCount{$tempAcc} > 1 and $skipDuplicates);
					$nameLabel{$tempAcc} = $tempName;
					$sequence{$tempAcc} = uc($tempSeq);
						$seqType = "protein" if ($tempSeq =~ /E/i or $tempSeq =~ /Q/i or $tempSeq =~ /I/i or $tempSeq =~ /L/i or $tempSeq =~ /F/i or $tempSeq =~ /P/i);
					}
				if ($lineRead =~ /^Frame=(\d)/)	# Correct for reading frame
					{
					my $readingFrame = $1;
					$sequence{$tempAcc} = "--" . $sequence{$tempAcc} if ($readingFrame == 2);
					$sequence{$tempAcc} = "-" . $sequence{$tempAcc} if ($readingFrame == 3);
					}
				}

			if ($inputType eq "phylip")
				{
				if ($lineRead =~ /^\s*(\d+)\s+(\d+)/)
					{
					$seqCount = $phylipTaxa = $1;
					$phylipChars = $2;
					$phylipLineCount = 0;
					}
				else
					{
					$phylipLineCount++;
					
					$lineRead =~ s/\s//g;
					
					$phylipSeq{$phylipLineCount} .= $lineRead;
					
					$phylipLineCount = 0 if ($phylipLineCount == $phylipTaxa);
					}
				}

			if ($inputType eq "GenBank")
				{
				# Get species name and accession number
					# Pure GenBank format
						$gbAcc = $1 if ($lineRead =~ /^\s*ACCESSION\s+(\w+\d+)/);
						if ($lineRead =~ /^\s*ORGANISM\s+/)
							{
							$seqCount++;
								$gbAcc = "tAlign_" . $seqCount if (not defined $gbAcc);
							$lineRead =~ s/^\s+//;
							my @orgLine = split(/\s+/, $lineRead);
								my $header = shift (@orgLine);
								$nameLabel{$gbAcc} = join('_', @orgLine);
								$accCount{$gbAcc}++;
							}
					# BioEdit format
						if ($lineRead =~ /^\s*TITLE/ and not defined ($gbAcc))
							{
							$seqCount++;
								$gbAcc = "tAlign_" . $seqCount;
							my @accLine = split (/\s+/, $lineRead);
								$gbAcc = $1 if ($accLine[2] =~ /^\((\w+\d+)\)/);
							$nameLabel{$gbAcc} = $accLine[1];
							$accCount{$gbAcc}++;
							}

				if ($lineRead =~ /^\s*ORIGIN/)
					{
					if ($accCount{$gbAcc} > 1)
						{
						print "\nWARNING: Accession number $gbAcc used more than once";
						if ($skipDuplicates)
							{
							print "; skipping this and subsequent entries\n";
							next;
							}
						else
							{
							print "; assigning temporary accession number\n";
							$gbAcc = "tAlign_" . $seqCount;
							$gbRead = 1;
							}
						}
					else
						{
						$gbRead = 1;
						$seqCount++;
						}
					next;
					}

				if ($lineRead =~ /^\s*\/\//)	# End of accession; process
					{
					$gbRead = 0;
					next unless ($gbAcc);
					
					push @accNum, $gbAcc unless ($accCount{$gbAcc} > 1);
					$geneticCode{$gbAcc} = $globalGenCode;
					$nameLabel{$gbAcc} =~ s/\s+/_/g;
						$nameLabel{$gbAcc} =~ s/\_+/_/g;
					$sequence{$gbAcc} =~ s/\d//g;
						$sequence{$gbAcc} =~ s/\s+//g;
						$sequence{$gbAcc} =~ s/\~/-/g if ($gapChange);
					undef $gbAcc;
					}
					
				next unless ($gbRead);
				
				if ($gbRead and $lineRead =~ /^\s+\d+/)
					{
					$sequence{$gbAcc} .= uc($lineRead);
					}
				}
			}
	close SEQ;
	
	if ($inputType eq "phylip")	# Postprocess input to derive taxon names and sequence; accounts for both sequential and extended formatting
		{
		for (my $i = 1; $i <= $phylipTaxa; $i++)
			{
			my $phylipAcc = "tAlign_" . $i;
			
			push @accNum, $phylipAcc;
			$geneticCode{$phylipAcc} = $globalGenCode;
			
			# Derive taxon name and sequence
				$sequence{$phylipAcc} = uc(substr($phylipSeq{$i}, 0 - $phylipChars));
					$seqType = "protein" if ($sequence{$phylipAcc} =~ /E/i or $sequence{$phylipAcc} =~ /Q/i or $sequence{$phylipAcc} =~ /I/i or $sequence{$phylipAcc} =~ /L/i or $sequence{$phylipAcc} =~ /F/i or $sequence{$phylipAcc} =~ /P/i);

				$nameLabel{$phylipAcc} = substr($phylipSeq{$i}, 0, length($phylipSeq{$i}) - $phylipChars);
					$nameLabel{$phylipAcc} =~ s/\s+/_/g;
					$nameLabel{$phylipAcc} =~ s/\_+/_/g;
			}
		}
	}

sub duplStrip	# Subroutine to strip duplicate entries from an array
	{
	my @duplArray = @_;
	my %dupl;		# Remove duplicate taxa
		my @stripArray = grep(!$dupl{$_}++, @duplArray);
		@stripArray = sort {$a cmp $b} @stripArray;
	return @stripArray;
	}

sub binMaker
	{
	$binWidth = ($binMaximum - $binMinimum) / $numBins;
	
	# Assign upper and lower boundaries
		for (my $i = 1; $i <= $numBins; $i++)
			{
			$binCount{$i} = 0;
			
			$lowerBin{$i} = 1 - ($i * $binWidth);
			$upperBin{$i} = 1 - (($i - 1) * $binWidth);
			}
	}

sub listMaker
	{
	my @list = duplStrip(@{$_[0]});
		@list = sort {$a <=> $b} @list;
	my $sortedList = "";
	
	for (my $i = 0; $i < scalar(@list); $i++)
		{
		if ($i == 0)
			{
			$sortedList .= $list[$i];
			}
		elsif ($i == scalar(@list) - 1)
			{
			$sortedList .= " " . $list[$i];
			}
		elsif ($list[$i] - $list[$i-1] == 1 and $list[$i+1] - $list[$i] == 1)
			{
			$sortedList .= "-";
			}
		else
			{
			$sortedList .= " " . $list[$i];
			}
		}

	$sortedList =~ s/\-+\s*/\-/g;
	return $sortedList;
	}

sub subset	# By Skeeve @ https://www.perlmonks.org/?node_id=259263; usage subset(\@refArray, \@testArray); returns 1 if subset
	{
	my ($refArray, $testArray) = @_;
    my (%subset);
    @subset{@$testArray} = @$testArray;
    delete @subset{@$refArray};
    return scalar(keys %subset) == 0;
	}

# Version history
#
# 	v1.0 (May 1, 2020)
#		- initial release
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is available at
# http://www.gnu.org/copyleft/gpl.html or by writing to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Fifth Floor, Boston, MA, 02110-1301, USA.
